import numpy as np
from scipy.interpolate import interp1d
from pathlib import Path

dets = ['ET','CE','aLIGO','aVirgo','KAGRA']

#------------------------------------------------------
# DETECTORS SENSITIVITY
#------------------------------------------------------

path = Path(__file__).parent / '../Detectors_Sensitivity/'

data_LIGO  = np.loadtxt(path / 'Sn_L.txt')
data_Virgo = np.loadtxt(path / 'Sn_V.txt')
data_Kagra = np.loadtxt(path / 'Sn_K.txt')
data_ET    = np.loadtxt(path / 'Sn_ET.txt')
data_CE    = np.loadtxt(path / 'Sn_CE.txt')

freq1, Sn1	= data_LIGO[:,0]  , data_LIGO[:,1]**2
freq2, Sn2	= data_Virgo[:,0] , data_Virgo[:,1]**2
freq3, Sn3	= data_Kagra[:,0] , data_Kagra[:,1]**2
freq4, Sn4	= data_ET[:,0]	  , data_ET[:,1]**2
freq5, Sn5	= data_CE[:,0]	  , data_CE[:,1]**2

freq1, Sn1	= data_LIGO[:,0]  , data_LIGO[:,1]**2
freq2, Sn2	= data_Virgo[:,0] , data_Virgo[:,1]**2
freq3, Sn3	= data_Kagra[:,0] , data_Kagra[:,1]**2
freq4, Sn4	= data_ET[:,0]	  , data_ET[:,1]**2
freq5, Sn5	= data_CE[:,0]	  , data_CE[:,1]**2

func1 = interp1d(freq1, Sn1, fill_value=np.inf, bounds_error=False)
func2 = interp1d(freq2, Sn2, fill_value=np.inf, bounds_error=False)
func3 = interp1d(freq3, Sn3, fill_value=np.inf, bounds_error=False)
func4 = interp1d(freq4, Sn4, fill_value=np.inf, bounds_error=False)
func5 = interp1d(freq5, Sn5, fill_value=np.inf, bounds_error=False)

#------------------------------------------------------

freq0 = 10**np.linspace(0,4,4000)

SnL 	   = func1(freq0)
SnV, SnK   = func2(freq0), func3(freq0)
SnET, SnCE = func4(freq0), func5(freq0)

PSD = {
'aLIGO': SnL,
'aVirgo': SnV,
'KAGRA': SnK,
'CE': SnCE,
'ET': SnET
}

#------------------------------------------------------
# LATEX LABELS
#------------------------------------------------------

labels_tex = {} 
labels_tex['m1']       = '$m_1$ [$M_{\odot}$]'
labels_tex['m2']       = '$m_2$ [$M_{\odot}$]'
labels_tex['Mc']       = '$M_c$ [$M_{\odot}$]'
labels_tex['eta']      = '$\eta$'
labels_tex['q']        = '$q$'
labels_tex['DL']       = '$d_L$ [Gpc]'
labels_tex['iota']     = '$\iota$ [rad]'
labels_tex['psi']      = '$\psi$ [rad]'
labels_tex['alpha']    = '$\\alpha$ [rad]'
labels_tex['beta']     = '$\\beta$ [rad]'
labels_tex['t_coal']   = '$t_{coal}$ [sec]'
labels_tex['phi_coal'] = '$\phi_{coal}$ [rad]'
labels_tex['sx1']      = '$S_{x,1}$'
labels_tex['sy1']      = '$S_{y,1}$'
labels_tex['sz1']      = '$S_{z,1}$'
labels_tex['sx2']      = '$S_{x,2}$'
labels_tex['sy2']      = '$S_{y,2}$'
labels_tex['sz2']      = '$S_{z,2}$'

labels_tex['RA']  = '$RA$ [deg]'
labels_tex['Dec'] = '$Dec$ [deg]'

def Load_Dictionaries():
	return PSD, labels_tex

#------------------------------------------------------


import numpy as np
import GWDALI.lib.Waveforms as wf
import GWDALI.lib.Angles_lib as geo
import GWDALI.lib.Dictionaries as gwdict
from scipy.interpolate import interp1d
from scipy.integrate import trapezoid
from scipy.optimize import root

rad = np.pi/180
deg = 1./rad

c = 299792458. # m/s
R_earth = 6371.e3 # meters

PSD, labels_tex = gwdict.Load_Dictionaries()

#----------------------------------------------
# m1(m2,Mc) or m2(m1,Mc)
def func_mMc(x,args):
	m, Mc = args
	return x**3 - (Mc**5/m**3)*x - Mc**5/m**2
def m_Mc(m,Mc,x0):
	res = root(func_mMc,x0,args=[m,Mc])
	return float(res.x[0])
#----------------------------------------------
# m1(eta,Mc) and m2(eta,Mc)
def eta_Mc(eta,Mc):
	m1 = 0.5*(Mc/eta**(3./5)) * (1 + np.sqrt(1-4*eta))
	m2 = 0.5*(Mc/eta**(3./5)) * (1 - np.sqrt(1-4*eta))
	return m1, m2
#----------------------------------------------
def q_Mc(q,Mc):
	m1 = ((1+q)/q**3)**(1./5)*Mc
	m2 = (q**2*(1+q))**(1./5)*Mc
	return m1, m2
#----------------------------------------------
def eta_m2(eta,m2): # return m1
	return 0.5*(m2/eta)*( (1-2*eta) + np.sqrt(1-4*eta))
def eta_m1(eta,m1): # return m2
	return 0.5*(m1/eta)*( (1-2*eta) - np.sqrt(1-4*eta))
#----------------------------------------------
def get_iota(prms):
	if("cos_iota" in prms.keys() ):
		iota = np.arccos(prms['cos_iota']) # rad
	elif("iota" in prms.keys() ):
		iota = prms['iota'] # rad
	else:
		print(">> Parameter 'iota' not found!")
		print(">> Received Parameters:", list(prms.keys()))
		quit()
	return iota

def get_dL(prms):
	if("ln_dL" in prms.keys()):
		dL = np.exp( prms['ln_dL'] ) * 1.e3
	elif("inv_dL" in prms.keys()):
		dL = 1.e3/prms['inv_dL']
	elif("DL" in prms.keys()):
		dL = prms["DL"] * 1.e3
	else:
		print(">> Parameter 'DL' not found!")
		print(">> Received Parameters:", list(prms.keys()))
		quit()
	return dL

def get_spins(prms):
	if(all([x in prms.keys() for x in ['sx1','sy1','sz1','sx2','sy2','sz2'] ])):
		sx1 = prms['sx1']
		sy1 = prms['sy1']
		sz1 = prms['sz1']
		sx2 = prms['sx2']
		sy2 = prms['sy2']
		sz2 = prms['sz2']
	elif( all( [x in prms.keys() for x in ['S1','theta_1','phi_1','S2','theta_2','phi_2']] ) ):
		sx1 = prms['S1']*np.sin( prms['theta_1'] )*np.cos( prms['phi_1'] )
		sy1 = prms['S1']*np.sin( prms['theta_1'] )*np.sin( prms['phi_1'] )
		sz1 = prms['S1']*np.cos( prms['theta_1'] )
		sx2 = prms['S2']*np.sin( prms['theta_2'] )*np.cos( prms['phi_2'] )
		sy2 = prms['S2']*np.sin( prms['theta_2'] )*np.sin( prms['phi_2'] )
		sz2 = prms['S2']*np.cos( prms['theta_2'] )
	else:
		print(">> Invalid or Incomplete Spin Inputs!")
		print(">> GWDALI works with:")
		print("\t ['S1','theta_1','phi_1','S2','theta_2','phi_2'] , or")
		print("\t ['sx1','sy1','sz1','sx2','sy2','sz2']")
		quit()
	return [sx1, sy1, sz1], [sx2, sy2, sz2]

def get_mass(prms):
	keys = list(prms.keys())
	Mass = {}	
	for key in keys:
		if( key in ['m1','m2','q','eta','Mc'] ):
			Mass[key] = prms[key]
	
	keys = Mass.keys()
	if(all(x in ['m1','m2'] for x in keys) ):
		return Mass['m1'], Mass['m2']
	elif(all(x in ['eta','Mc'] for x in keys) ):
		eta, Mc = Mass['eta'], Mass['Mc']
		return eta_Mc(eta,Mc)
	elif(all(x in ['eta','m1'] for x in keys) ):
		m1, eta = Mass['m1'], Mass['eta']
		m2 = eta_m1(eta,m1)
		return m1, m2
	elif(all(x in ['eta','m2'] for x in keys) ):
		m2, eta = Mass['m2'], Mass['eta']
		m1 = eta_m2(eta,m2)
		return m1, m2
	elif( all(x in ['m1','q'] for x in keys) ):
		m1, q = Mass['m1'], Mass['q']
		m2 = m1*q
		return m1, m2
	elif( all(x in ['m2','q'] for x in keys) ):
		m2, q = Mass['m2'], Mass['q']
		m1 = m2/q
		return m1, m2
	elif( all(x in ['q','Mc'] for x in keys) ):
		q, Mc = Mass['q'], Mass['Mc']
		return q_Mc(q,Mc)
	elif( all(x in ['m1','Mc'] for x in keys) ):
		m1, Mc = Mass['m1'], Mass['Mc']
		m2 = m_Mc(m1,Mc,m1)
		return m1, m2
	elif( all(x in ['m2','Mc'] for x in keys) ):
		m2, Mc = Mass['m2'], Mass['Mc']
		m1 = m_Mc(m2,Mc,m2)
		return m1, m2
	if(all(x in ['q','eta'] for x in keys) ):
		print("---"*10)
		print("Error! It is not possible to recover (m1,m2) from (q,Mc)")
		print("---"*10) ; quit()
	else:
		print("---"*10)
		print("Error! We need at least 2 mass parameters!")
		print("Parameters allowed: (m1, m2, q, eta, Mc)")
		print("---"*10) ; quit()

#------------------------------------------------

def Pattern_Func(alpha,beta,psi,Omega):
	u = np.cos(beta) ; Coeff = np.sin(Omega)
	fp = 0.5*(1.+u**2)*np.cos(2.*alpha)*Coeff
	fx = -u*np.sin(2.*alpha)*Coeff
	
	Fp =  fp*np.cos(2.*psi) + fx*np.sin(2.*psi)
	Fx = -fp*np.sin(2.*psi) + fx*np.cos(2.*psi)

	return Fp, Fx

#-------------------------------------------------

def ScalarProduct(freq,Sn,A,B):
	return 4*np.real( trapezoid( A*np.conj(B)/Sn, freq) )

def GW_Polarizations(params, freq, approx):
	keys = list(params.keys())
	
	m1, m2   = get_mass(params)
	iota 	 = get_iota(params) # rad
	DL       = get_dL(params)   # Mpc
	psi      = params['psi']    # rad

	s1, s2 = get_spins(params)

	hp, hx, freq0 = wf.Waveforms(m1,m2,iota,DL,s1,s2,freq, approx=approx)

	return hp, hx, freq0

# det: abreviation of detector
def Signal(params,dets,approx):
	alpha    = params['RA']*rad       # rad
	beta     = (90-params['Dec'])*rad # rad
	psi      = params['psi']          # rad
	t_coal   = params['t_coal']       # sec
	phi_coal = params['phi_coal']     # rad
	iota = get_iota(params) # rad

	det_a, det_b = dets
	tau_a    = geo.get_TimeDelay(det_a,params) # sec (time-delay between det_a and geocenter)
	tau_b    = geo.get_TimeDelay(det_b,params) # sec (time-delay between det_b and geocenter)
	tau_ab = (tau_a - tau_b) # time_delay (between det_a and det_b)

	name  = det_b['name']
	freq  = det_b['freq'].copy()
	lon   = det_b['lon'] # deg
	lat   = det_b['lat'] # deg
	rot   = det_b['rot'] # deg

	alpha_obs, beta_obs, psi_obs = geo.ObsAngles(alpha,beta,iota,psi,lon,lat,rot)
	Fp, Fx = Pattern_Func(alpha_obs,beta_obs,psi_obs,det_b['shape']*rad)
	hp, hx, freq0 = GW_Polarizations(params, freq, approx)

	# t_coal := t_(0,a)
	# tau_ab := tau_a - tau_b = t_(0,a) - t_(0,b)
	# t_(0,b) = t_(0,a) - tau_ab
	Phase = 2*np.pi*freq0*(t_coal-tau_ab) - phi_coal
	H = (Fp*hp + Fx*hx)*np.exp(1.j*Phase)
	
	gw_signal = interp1d(freq0,H,bounds_error=False,fill_value='extrapolate')
	H = gw_signal(freq)

	return H

#-------------------------------------------------

# eps (standard) = 1.e-6
def split_prms(params,x,eps, diff_order):
	p = params[x]
	dx = np.max([eps,eps*p])
	P0 = params.copy() ; P0[x] = p - dx	
	P1 = params.copy() ; P1[x] = p - dx/2
	P2 = params.copy() ; P2[x] = p + dx/2
	P3 = params.copy() ; P3[x] = p + dx
	if(diff_order == 2): return [P1,P2], dx
	elif(diff_order==4): return [P0,P1,P2,P3], dx
	else:
		print("\n\t Invalid diff_order! Allowed values: [2,4] ")
		quit()

def Diff1(x, params, dets, approx, eps, diff_order):
	Ps, dx = split_prms(params,x,eps, diff_order)
	Y = [Signal(P, dets, approx) for P in Ps]
	if(diff_order == 2): return (Y[1]-Y[0])/dx
	elif(diff_order==4): return 4*(Y[2]-Y[1])/(3*dx) - (Y[3]-Y[0])/(6*dx)

def Diff2(xi, xj, params, dets, approx, eps, diff_order):
	Ps, dx = split_prms(params,xi,eps, diff_order)
	Y = [Diff1(xj, P, dets, approx, eps, diff_order) for P in Ps]
	if(diff_order == 2): return (Y[1]-Y[0])/dx
	elif(diff_order==4): return 4*(Y[2]-Y[1])/(3*dx) - (Y[3]-Y[0])/(6*dx)

def Diff3(xi, xj, xk, params, dets, approx, eps, diff_order):
	Ps, dx = split_prms(params,xi,eps, diff_order)
	Y = [Diff2(xj,xk, P, dets, approx, eps, diff_order) for P in Ps]
	if(diff_order == 2): return (Y[1]-Y[0])/dx
	elif(diff_order==4): return 4*(Y[2]-Y[1])/(3*dx) - (Y[3]-Y[0])/(6*dx)

#-------------------------------------------------#-------------------------------------------------
#-------------------------------------------------#-------------------------------------------------

def Fisher_ij(xi,xj, params, dets, approx, eps, diff_order): # [1,1]
	det = dets[1]
	Dxi = Diff1(xi, params, dets, approx, eps, diff_order)
	Dxj = Diff1(xj, params, dets, approx, eps, diff_order)
	return ScalarProduct(det['freq'], det['Sn'], Dxi, Dxj )

#-------------------------------------------------#-------------------------------------------------
# (arXiv:2203.02670)

def Doublet12(xi,xj,xk, params, dets, approx, eps, diff_order): # [1,2]
	det = dets[1]
	D_i  = Diff1(xi, params, dets, approx, eps, diff_order)
	D_jk = Diff2(xj,xk, params, dets, approx, eps, diff_order)
	return ScalarProduct(det['freq'], det['Sn'], D_i, D_jk)

def Doublet22(xi,xj,xk,xl, params, dets, approx, eps, diff_order): # [2,2]
	det = dets[1]
	D_ij = Diff2(xi,xj, params, dets, approx, eps, diff_order)
	D_kl = Diff2(xk,xl, params, dets, approx, eps, diff_order)
	return ScalarProduct(det['freq'], det['Sn'], D_ij, D_kl)

#-------------------------------------------------#-------------------------------------------------

def Triplet13(xi,xj,xk,xl, params, dets, approx, eps, diff_order): # [1,3]
	det = dets[1]
	D_i   = Diff1(xi, params, dets, approx, eps, diff_order)
	D_jkl = Diff3(xj,xk,xl, params, dets, approx, eps, diff_order)
	return ScalarProduct(det['freq'], det['Sn'], D_i, D_jkl)

def Triplet23(xi,xj,xk,xl,xm, params, dets, approx, eps, diff_order): # [2,3]
	det = dets[1]
	D_ij  = Diff2(xi,xj, params, dets, approx, eps, diff_order)
	D_klm = Diff3(xk,xl,xm, params, dets, approx, eps, diff_order)
	return ScalarProduct(det['freq'], det['Sn'], D_ij, D_klm)

def Triplet33(xi,xj,xk,xl,xm,xn, params, dets, approx, eps, diff_order): # [3,3]
	det = dets[1]
	D_ijk = Diff3(xi,xj,xk, params, dets, approx, eps, diff_order)
	D_lmn = Diff3(xl,xm,xn, params, dets, approx, eps, diff_order)
	return ScalarProduct(det['freq'], det['Sn'], D_ijk, D_lmn)

#-------------------------------------------------#-------------------------------------------------



import os, sys
import numpy as np
import GWDALI.lib.Dictionaries as dct
import GWDALI.lib.Diff_Signal_Tensors as gw
import GWDALI.lib.Auxiliar as sym

from itertools import permutations
from tqdm import trange

deg2rad = np.pi/180
c = 299792458. # m/s
R_earth = 6371.e3 # meters

PSD, labels_tex = dct.Load_Dictionaries()

def pmt(vec):
	x, y = [ list(xx) for xx in list(permutations(vec)) ] , []
	for a in x:
		if(not a in y): y.append(a)
	return y

def Get_Tensors(FreeParams, gw_prms, detectors, dali_method, approx, step_size, diff_order, hide_info):
	
	Np = len(FreeParams)
	rho2 = 0

	Fisher = np.zeros([Np, Np])
	Db12    = np.zeros([Np, Np, Np])
	Db22    = np.zeros([Np, Np, Np, Np])
	Tp13    = np.zeros([Np, Np, Np, Np])
	Tp23    = np.zeros([Np, Np, Np, Np, Np])
	Tp33    = np.zeros([Np, Np, Np, Np, Np, Np])

	num_det = 1
	GwData = [] ; SNR = []
	det_a = detectors[0]
	for det_b in detectors:
		dets = [det_a, det_b]
		#-----------------------------------
		# Computing Signal to Noise 
		h 	  = gw.Signal(gw_prms, dets, approx)#*ExpT

		SNR2  = gw.ScalarProduct(det_b['freq'], det_b['Sn'],h,h)
		rho2 += SNR2
		GwData.append(h) ; SNR.append(np.sqrt(SNR2))
		#------------------------------------------------------
		# Computing Fisher Matrix
		#------------------------------------------------------

		if(not hide_info): print("\n\n\t Computing Fisher %s(%d) ..." % (det_b['name'],num_det))
		for i in range(Np):
			for j in range(i+1):
				if(not hide_info): print('Fisher: [%d-%d]' % (i+1,j+1), end='\r')
				xj = FreeParams[j]
				xi = FreeParams[i]
				Fij = gw.Fisher_ij(xi, xj, gw_prms, dets, approx, step_size, diff_order)
				Fisher[i][j] += Fij

				if(i!=j): Fisher[j][i] += Fij
		if(not hide_info): print("\n\n\t\tFisher Concluded!!!\n\n")
		#------------------------------------------------------
		# Computing Doublet/Triplet (arXiv:2203.02670)
		#------------------------------------------------------
		
		idxs_doub, idxs_trip = sym.get_indep_indexies(Np)
		
		if(dali_method in ["Doublet","Triplet"]):
			indepD_12, indepD_22 = idxs_doub
			# --------------Doublet(1,2)--------------		
			for Idx2 in indepD_12:
				i, j, k = Idx2
				value = gw.Doublet12(*[FreeParams[ii] for ii in Idx2],gw_prms,dets,approx, step_size, diff_order)
				for p in pmt([j,k]):
					jj, kk = p
					Db12[i][jj][kk] += value
			# --------------Doublet(2,2)--------------
			for Idx2 in indepD_22:
				i, j, k, l = Idx2
				value = gw.Doublet22(*[FreeParams[ii] for ii in Idx2],gw_prms,dets,approx, step_size, diff_order)
				for p1 in pmt([i,j]):
					ii, jj = p1
					for p2 in pmt([k,l]):
						kk, ll = p2
						Db22[ii][jj][kk][ll] += value
						if(not hide_info): print('(%s) [Sym] Doublet: '%det_b['name'], i, j, k, l,'...'*10, end='\r')
			#*****************************#*****************************#*****************************
			if(dali_method == 'Triplet'):
				indepT_13, indepT_23, indepT_33 = idxs_trip
				# --------------Triplet(1,3)--------------
				for Idx3 in indepT_13:
					i, j, k, l = Idx3
					value = gw.Triplet13(*[FreeParams[ii] for ii in Idx3],gw_prms,dets,approx, step_size, diff_order)
					for p in pmt([j,k,l]):
						jj, kk, ll = p
						Tp13[i][jj][kk][ll] += value
				# --------------Triplet(2,3)--------------
				for Idx3 in indepT_23:
					i, j, k, l, m = Idx3
					value = gw.Triplet23(*[FreeParams[ii] for ii in Idx3],gw_prms,dets,approx, step_size, diff_order)
					for p1 in pmt([i,j]):
						ii, jj = p1
						for p2 in pmt([k,l,m]):
							kk, ll, mm = p2
							Tp23[ii][jj][kk][ll][mm] += value
				# --------------Triplet(3,3)--------------
				for Idx3 in indepT_33:
					i, j, k, l, m, n = Idx3
					value = gw.Triplet33(*[FreeParams[ii] for ii in Idx3],gw_prms,dets,approx, step_size, diff_order)
					for p1 in pmt([i,j,k]):
						ii, jj, kk = p1
						for p2 in pmt([l,m,n]):
							ll, mm, nn = p2
							Tp33[ii][jj][kk][ll][mm][nn] += value
							if(not hide_info): print('(%s) [Sym] Triplet: '%det_b['name'], i, j, k, l, m, n,'...'*10, end='\r')
		# --------------# --------------# --------------					

	SNR_sum = np.sqrt(rho2)
	if(not hide_info):
		print(" >> SNR = ", SNR_sum)
		print(" >> 100/SNR = %.2e" % (100/SNR_sum) + '%')

	Doublet = [Db12, Db22]
	Triplet = [Tp13, Tp23, Tp33]

	return [SNR, GwData, Fisher , Doublet, Triplet]


from setuptools import setup, find_packages

with open("README.md",'r') as arq:
	readme = arq.read()

setup(
	name = 'gwdali',
	version = '0.1.4',
	license = 'BSD 3-Clause License',
	author  = 'Josiel Mendonça Soares de Souza',
	long_description = readme,
	long_description_content_type = "text/markdown",
	author_email = 'josiel.jms7@gmail.com',
	keywords = 'fisher matrix, gravitational waves, gw, dali',
	description = 'A Fisher-Based Software for Parameter Estimation from Gravitational Waves',
	packages = find_packages(),
	include_package_data=True,
	package_data={
        'GWDALI': ['Detectors_Sensitivity/*.txt'],
    },
	install_requeries = ['numpy','matplotlib','scipy','bilby','astropy','itertools'],
	#url = "https://github.com/jmsdsouzaPhD/gwdali/",
)



import numpy as np
import matplotlib.pyplot as plt

def scl_prod(V1,V2): # Scalar Product
	if(len(V1)!=len(V2)):
		print("Error!")
		quit()
	SP = 0 ; ndim = len(V1)
	for i in range(ndim): SP += V1[i]*V2[i]
	return SP

def vec_prod(V1,V2): # Vectorial Product
	if(len(V1)!=len(V2)):
		print("Error!")
		quit()
	vPx = V1[1]*V2[2] - V1[2]*V2[1]
	vPy = V1[2]*V2[0] - V1[0]*V2[2]
	vPz = V1[0]*V2[1] - V1[1]*V2[0]
	return vPx, vPy, vPz

def AngTransf(alpha,beta,theta,phi,ksi): # Sky Position from Geocentric to Detector Coordinates
	sA   = np.sin(beta)*np.sin(alpha-phi)
	cA   = np.sin(beta)*np.cos(theta)*np.cos(alpha-phi) - np.cos(beta)*np.sin(theta)
	cosB = np.sin(beta)*np.sin(theta)*np.cos(alpha-phi) + np.cos(beta)*np.cos(theta)
	cosA =  cA*np.cos(ksi) + sA*np.sin(ksi)
	sinA = -cA*np.sin(ksi) + sA*np.cos(ksi)
	alpha_det = np.arctan2(sinA,cosA)
	beta_det = np.arccos(cosB)
	return alpha_det, beta_det

#==================================//==================================//==============

def Normal_Vec(theta,phi): # Detector position in Geocentric Coordinates
	Zx = np.sin(theta)*np.cos(phi)
	Zy = np.sin(theta)*np.sin(phi)
	Zz = np.cos(theta)
	return Zx, Zy, Zz

def Nvec(alpha,beta,theta,phi,ksi): # Sky position in Detector Coordinates
	alpha_det, beta_det = AngTransf(alpha,beta,theta,phi,ksi)
	Nx = np.sin(beta_det)*np.cos(alpha_det)
	Ny = np.sin(beta_det)*np.sin(alpha_det)
	Nz = np.cos(beta_det)
	return Nx, Ny, Nz 

def Lvec0(alpha,beta,iota,psi): # Compute L in Geocentric Coordinates
	Lx = -np.sin(iota)*np.sin(psi)*np.cos(beta)*np.cos(alpha) - np.sin(iota)*np.cos(psi)*np.sin(alpha) + np.cos(iota)*np.sin(beta)*np.cos(alpha)
	Ly = -np.sin(iota)*np.sin(psi)*np.cos(beta)*np.sin(alpha) + np.sin(iota)*np.cos(psi)*np.cos(alpha) + np.cos(iota)*np.sin(beta)*np.sin(alpha)
	Lz = np.sin(iota)*np.sin(psi)*np.sin(beta) + np.cos(iota)*np.cos(beta)
	return Lx, Ly, Lz

def Lvec(L0,theta,phi,ksi): # Compute L in Detector Coordinates
	Lx0, Ly0, Lz0 = L0
	r11 = np.cos(theta)*np.cos(phi)*np.cos(ksi)-np.sin(phi)*np.sin(ksi)
	r12 = np.cos(theta)*np.sin(phi)*np.cos(ksi) + np.cos(phi)*np.sin(ksi)
	r13 = -np.sin(theta)*np.cos(ksi)
	r21 = -np.cos(theta)*np.cos(phi)*np.sin(ksi) - np.sin(phi)*np.cos(ksi)
	r22 = -np.cos(theta)*np.sin(phi)*np.sin(ksi) + np.cos(phi)*np.cos(ksi)
	r23 = np.sin(theta)*np.sin(ksi)
	r31 = np.sin(theta)*np.cos(phi)
	r32 = np.sin(theta)*np.sin(phi)
	r33 = np.cos(theta)
	Lxi = r11*Lx0 + r12*Ly0 + r13*Lz0
	Lyi = r21*Lx0 + r22*Ly0 + r23*Lz0
	Lzi = r31*Lx0 + r32*Ly0 + r33*Lz0
	return Lxi, Lyi, Lzi

def poll_ang(alpha0,beta0,iota,psi0,theta,phi,ksi):
	N = Normal_Vec(beta0,alpha0)
	Z = Normal_Vec(theta,phi)
	L = Lvec0(alpha0,beta0,iota,psi0)

	sPsi = scl_prod(Z,L) - scl_prod(N,L)*scl_prod(N,Z)
	cPsi = scl_prod(Z,vec_prod(N,L))
	psi = np.arctan2(sPsi,cPsi)
	return psi

def ObsAngles(alpha0,beta0,iota,psi0,lon,lat,rot):
	phi   = lon*np.pi/180
	theta = (90-lat)*np.pi/180
	ksi   = rot*np.pi/180

	alpha_obs, beta_obs = AngTransf(alpha0,beta0,theta,phi,ksi)
	psi_obs = poll_ang(alpha0,beta0,iota,psi0,theta,phi,ksi)
	return alpha_obs, beta_obs, psi_obs

rad = np.pi/180
R_earth = 6371.e3
c = 299792458.

def get_TimeDelay(det,prms):
	alpha0 = prms['RA']*rad
	beta0  = (90-prms['Dec'])*rad
	phi    = det['lon']*rad
	theta  = (90-det['lat'])*rad
	ksi    = det['rot']*rad

	_, beta_a = AngTransf(alpha0,beta0,theta,phi,ksi)
	return np.cos(beta_a)*R_earth/c